package index;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.mortbay.util.MultiMap;

import window.ExractFingerprints;
import window.FileTooShort;
import window.Fingerprint;
import window.KGramOutOfBound;

/**
 * This will get the directory which contains the file and make the collection from them.
 * Bulding Inveted index and Index of the documents. 
 */
public class Thesaurus {
	/**
	 * Inverted index of the collection
	 */
	private MultiMap invertedIndex;
	
	/**
	 * Index of the documents
	 */
	private Map allFilesFingerprints;
	
	/**
	 * List of the files for obtaining information about files
	 */
	private File[] files;
	
	/**
	 * List of fileIDs. it's another view of allFilesFingerprints
	 */
	private FileID[] fileIDs;

	/**
	 * Make the thesaurus from the files in the specified directory
	 * 
	 * @param path the directory containing the files of the collection
	 * @throws DirectoryNotFound the directory must exist
	 * @throws FileTooShort
	 */
	public Thesaurus(String path) throws DirectoryNotFound, FileTooShort {
		File root = new File( path);
		
		if(!root.isDirectory())
			throw new DirectoryNotFound();
		
		files =  root.listFiles();

		allFilesFingerprints = getFingerprints();
		invertedIndex = makeInvertedIndex();
		fileIDs = fillFileIDs(); 
	}
	
	//TODO Set
	/**
	 * Getting the terms of the specified document name 
	 * @param document the name of the document
	 * @return the list of its terms
	 */
	public List getDocumentTerms(String document) {
		if(!allFilesFingerprints.containsKey(document))
			return null;
		
		ArrayList retval = new ArrayList();
		Fingerprint[] fingerprints = (Fingerprint[])allFilesFingerprints.get(document);
		
		for (int i = 0; i < fingerprints.length; i++) {
			Term t = new Term(fingerprints[i]);
			
			if(retval.contains(t))
				((Term)retval.get(retval.indexOf(t))).incFrequency();
			else
				retval.add(t);
		}
		
		return retval;
	}
	
	/**
	 * Getting the number of collection documents.
	 * @return the number of the collection documents
	 */
	public int getNumberOfDocuments() {
		return files.length;
	}
	
	/**
	 * Getting the number of the documents containing the specific gram(fingerprint)
	 * @param gram the gram which number of its occurrences must be computed
	 * @return the number of occurrences
	 */
	public int getNumberOfDocumentsContainingGram(Integer gram) {
		return getDocuments(gram).size();
	}
	
	//TODO set
	/**
	 * The list of documents which contain the specific gram 
	 * @param gram the specified gram
	 * @return the list of documents
	 */
	public List getDocuments(Integer gram) {
		if(!invertedIndex.containsKey(gram))
			return null;
		
		List entry = invertedIndex.getValues(gram);
		List simple = new ArrayList();
		
		for (Iterator iter = entry.iterator(); iter.hasNext();) {
			IndexEntry element = (IndexEntry) iter.next();
			
			if(!simple.contains(element.getFileName()))
				simple.add((element.getFileName()));
		}
		
		return simple;
	}
	
	/**
	 * Make another view of allFilesFingerprints in the form of FileID[]
	 * @return another view of allFilesFingerprints
	 */
	private FileID[] fillFileIDs() {
		FileID[] retval = new FileID[allFilesFingerprints.size()];
		
		for (int i = 0; i < files.length; i++) {
			String fileName = files[i].getName();
			
			retval[i] = new FileID(fileName, (Fingerprint[])allFilesFingerprints.get(fileName));
		}
		
		return retval;
	}
	
//	private FileID[] fillFileIDs() {
//		FileID[] retval = new FileID[allFilesFingerprints.size()];
//
//		int i = 0;
//		
//		for (Iterator it = allFilesFingerprints.keySet().iterator(); it.hasNext();i++) {
//			String fileName = (String) it.next();
//			
//			retval[i] = new FileID(fileName, (Fingerprint[])allFilesFingerprints.get(fileName));
//		}
//		
//		return retval;
//	}

	/**
	 * Make the inverted index 
	 * @return inverted index
	 */
	private MultiMap makeInvertedIndex() {
		MultiMap retval = new MultiMap();
		
		SortedSet keys = new TreeSet( allFilesFingerprints.keySet());
		
		for (Iterator it = keys.iterator(); it.hasNext();) {
			String fileName = (String) it.next();
			Fingerprint[] fileFingerprints = (Fingerprint[])allFilesFingerprints.get(fileName);
			
			for (int j = 0; j < fileFingerprints.length; j++) {
				Fingerprint fp = fileFingerprints[j];
				
				retval.add( new Integer( fp.getValue()), new IndexEntry(fileName, fp));
			}
		}
		
		return retval;
	}
	
	/**
	 * Make the index from the documents
	 * @return the index
	 * @throws FileTooShort
	 */
	private Map getFingerprints() throws FileTooShort {
		ArrayList fingerprintsAL = extractAllGrams( files);
		
		HashMap retval = new HashMap();
		int i = 0;
		
		for (Iterator it = fingerprintsAL.iterator(); it.hasNext(); i++) {
			ArrayList fileFingerprintsAL = (ArrayList) it.next();
			Fingerprint[] fileFingerprints = (Fingerprint[])fileFingerprintsAL.toArray(new Fingerprint[1]);
			
			retval.put(files[i].getName(), fileFingerprints);
		}
		
		return retval;
	}
	
	/**
	 * Each file has corresponding entry in for the list of the fingerprints in the return value 
	 * @param files the files to obtain fingerprints
	 * @return list of fingerprints
	 * @throws FileTooShort
	 */
	private ArrayList extractAllGrams(File[] files) throws FileTooShort {
		ArrayList retval = new ArrayList();
		
		for (int i = 0; i < files.length; i++) {
			try {
				ExractFingerprints fpse = new ExractFingerprints(files[i].toString()); 
				retval.add( fpse.extract());
			} catch (FileNotFoundException e) {
				System.out.println("File must be found if not deleted during running");
				e.printStackTrace(System.err);
				System.exit(1);
			} catch (KGramOutOfBound e) {
				System.out.println("Fatal: Must not be out of bound");
				e.printStackTrace(System.err);
				System.exit(1);
			}
		}
		
		return retval;
	}
	
	public Map getAllFilesFingerprints() {
		return allFilesFingerprints;
	}

	public FileID[] getFileIDs() {
		return fileIDs;
	}

	public MultiMap getInvertedIndex() {
		return invertedIndex;
	}
}
