package window;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;

/**
 * This is the heart of the application which will apply winnowing algorithm to one document 
 * and will extract fingerprints
 */
public class ExractFingerprints {
	/**
	 * The path to file
	 */
	private String path;
	
	/**
	 * The string representation of document
	 */
	private String content;
	
	/**
	 * The offset of the begining of the window
	 */
	private int offset;
	
	public ExractFingerprints(String path) throws FileNotFoundException, FileTooShort {
		this.path = path;
		fileToString();
		offset = 0;
	}
	
	/**
	 * The center and the heart of the application which roll the window and fingerprints
	 * @return the fingerprints of the document
	 * @throws KGramOutOfBound
	 */
	public ArrayList extract() throws KGramOutOfBound {
		//The window
		//its size cannot be less than 2
		int[] w = new int[CONSTANTS.W];
		//Right end
		int r = w.length - 1;
		int min = r;
		ArrayList fingerPrints = new ArrayList();

		for(int i = 0; i < w.length; i++)
			w[i] = nextHash();
		
		for(int i = w.length - 2; i != -1; i--) {
			if(w[i] < w[min])
				min = i;
		}
		
		fingerPrints.add( record( w[min], globalPosition(min, r)));
		
		while(hasNext()) {
			r = (r + 1) % w.length;
			w[r] = nextHash();
			
			if(min == r) {
				for(int i = (r - 1 + w.length) % w.length; i != r; i = (i - 1 + w.length) % w.length) {
					if(w[i] < w[min])
						min = i;
				}
				
				fingerPrints.add( record( w[min], globalPosition(min, r)));
			} else {
				if(w[r] < w[min]) {
					min = r;
					fingerPrints.add( record(w[min], globalPosition(min, r)));
				}
			}
		}
		
		Collections.sort(fingerPrints);
//		System.out.println(fingerPrints);
		
		return fingerPrints;
	}
	
	/**
	 * Convert file to string
	 * @throws FileNotFoundException file must exist
	 * @throws FileTooShort file must not be less than W + K - 1
	 */
	private void fileToString() throws FileNotFoundException, FileTooShort {
		BufferedReader read = new BufferedReader(new FileReader(path));

		try {
			content = read.readLine();
			
			while(read.ready())
				content += read.readLine();
			
			if( content == null)
				content = "";
			
			if( content.length() < CONSTANTS.W + CONSTANTS.K - 1) {
				System.err.println("The size of the " + path + " is " + content.length() + " it must att least be W + K - 1 which is: " + (CONSTANTS.W + CONSTANTS.K - 1));
				throw new FileTooShort();
			}
			
			read.close();
		} catch (IOException e) {
			System.out.println("Unable to operate on file: " + path);
			
			e.printStackTrace(System.err);
		}
	}
	
	/**
	 * Determine if the window can be slipped more to he right
	 * @return the condition
	 */
	private boolean hasNext() {
		return (offset + CONSTANTS.K - 1 <= content.length() - 1); 
	}
	
	/**
	 * Getting the next gram
	 * @return the next K-Gram
	 * @throws KGramOutOfBound there's no more gram
	 */
	private int nextHash() throws KGramOutOfBound {
		if(!hasNext())
			throw new KGramOutOfBound();
		
		String gram = content.substring(offset, offset + CONSTANTS.K);
		offset++;
		
//		System.out.println(gram + "\t" + gram.hashCode());
		
		return gram.hashCode();
	}
	
	/**
	 * Compute the global position for one gram
	 * @param min the offset of the gram
	 * @param rightEnd the offset of the right end of the window
	 * @return the global offset in the document
	 */
	private int globalPosition(int min, int rightEnd) {
		int position = offset;
		position -= ((rightEnd - min + CONSTANTS.W) % CONSTANTS.W);
		
		return position;
	}
	
	/**
	 * Save the fingerprint
	 * @param hashValue the int value corresponding to K-Gram
	 * @param position the place in the document
	 * @return the fingerprint
	 */
	private Fingerprint record(int hashValue, int position) {
		return new Fingerprint(hashValue, position, offset - CONSTANTS.W + 1);
	}
}
